from abc import ABCMeta, abstractmethod
from functools import cached_property
from typing import Optional

from ..utils import *


class Package(metaclass=ABCMeta):
    os: Optional[list[OS]]
    distribution: Optional[list[Distribution]]

    def __init__(
        self, os: Some[OS] = None, distribution: Some[Distribution] = None
    ) -> None:
        self.os = unsome_(os)
        self.distribution = unsome_(distribution)

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def get_remote_version(self) -> str:
        pass

    @cached_property
    def remote_version(self) -> str:
        return self.get_remote_version()

    @property
    @abstractmethod
    def local_version(self) -> Optional[str]:
        pass

    @property
    def is_installed(self) -> bool:
        return self.remote_version == self.local_version

    @abstractmethod
    def install(self) -> None:
        pass

    def ensure(self) -> None:
        if self.os is not None and CURRENT_OS not in self.os:
            return
        if (
            CURRENT_DISTRIBUTION is not None
            and self.distribution is not None
            and CURRENT_DISTRIBUTION not in self.distribution
        ):
            return
        if self.is_installed:
            return
        self.install()



from abc import ABCMeta

import requests

from .base import Package


class PipBasePackage(Package, metaclass=ABCMeta):
    package: str

    @property
    def name(self) -> str:
        return self.package

    def get_remote_version(self) -> str:
        pypi_info = requests.get(f"https://pypi.org/pypi/{self.package}/json").json()
        return pypi_info["info"]["version"]



import dataclasses
import os
import sqlite3
import threading
from functools import cached_property
from typing import Generic, Type, TypeVar

T = TypeVar("T")

DB_PATH = "state.sqlite"


DB_CACHE = threading.local()


def db() -> sqlite3.Connection:
    try:
        return DB_CACHE.connection
    except AttributeError:
        pass
    connection = sqlite3.connect(DB_PATH)
    connection.row_factory = sqlite3.Row
    DB_CACHE.connection = connection
    return connection


class Storage(Generic[T]):
    @cached_property
    def attributes(self) -> list[str]:
        return [field.name for field in dataclasses.fields(self.klass)]

    def __init__(self, name: str, klass: Type[T]) -> None:
        self.name = name
        self.klass = klass
        db().execute(
            f'CREATE TABLE IF NOT EXISTS {self.name} (id TEXT PRIMARY KEY, {", ".join(self.attributes)})'
        )

    def __getitem__(self, key: str) -> T:
        row = db().execute(f"SELECT * FROM {self.name} WHERE id = ?", (key,)).fetchone()
        if row:
            attributes = {key: row[key] for key in row.keys() if key != "id"}
            return self.klass(**attributes)
        raise KeyError(key)

    def __delitem__(self, key: str) -> None:
        db().execute(f"DELETE FROM {self.name} WHERE id = ?", (key,))

    def __setitem__(self, key: str, value: T) -> None:
        with db():
            del self[key]
            attr_clause = "?"
            attr_values = [key]
            for attribute in self.attributes:
                attr_clause += ", ?"
                attr_values.append(getattr(value, attribute))
            db().execute(f"INSERT INTO {self.name} VALUES ({attr_clause})", attr_values)



import os
from typing import Optional

from ..utils import run, run_output, with_os
from .base import Package


def get_current_linux_shell() -> str:
    return run_output("getent", "passwd", os.environ["USER"]).split(":")[6].strip()


def get_current_macos_shell() -> str:
    return run_output("dscl", ".", "-read", os.environ["HOME"], "UserShell").split(
        ": "
    )[1]


SHELLS_FILE = "/etc/shells"


def get_all_shells() -> list[str]:
    with open(SHELLS_FILE) as shells_file:
        return [shell.strip() for shell in shells_file]


class Shell(Package):
    def __init__(self, shell: str, **kwargs) -> None:
        self.shell = shell
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "_shell"

    def get_remote_version(self) -> str:
        return self.shell

    @property
    def local_version(self) -> Optional[str]:
        return with_os(linux=get_current_linux_shell, macos=get_current_macos_shell)()

    def install(self) -> None:
        if not os.path.exists(self.shell):
            raise ValueError(f"{self.shell} does not exist.")
        if not os.access(self.shell, os.X_OK):
            raise ValueError(f"{self.shell} is not executable.")
        if self.shell not in get_all_shells():
            run("sudo", "tee", "-a", SHELLS_FILE, input=self.shell.encode())
        run("chsh", "-s", self.shell)
